      subroutine eig(ssp,zz,rr,bb,trk,s,par,xinfo,rpath)

c  A subroutine to ray trace.  The code implements range dependence, including 
c  simple bottom and surface specular reflections, if desired. 
c
c  par has real parameters with flags and values required for the calculation:
c
c  start angle
c      par(1)=ang1
c  end angle
c      par(2)=ang2
c  number of rays to trace
c      par(3)=numrays
c  source depth
c      par(4)=z0
c  receiver depth
c      par(5)=zr
c  range
c      par(6)=r2

c  timefront or eigenray flag
c      par(7)=itimefront
c  yes, no ray tracking flag
c      par(8)=itemplate
c  number of rays in template
c      par(9)=nrays
c  save ray path flag
c      par(10)=iraysave
c  range increment to save ray paths
c      par(11)=dxsav
c  tolerance for missing the receiver depth
c      par(12)=zmiss
c  gross tolerance for missing the receiver depth
c      par(13)=zmiss_max

c  travel time offset
c      par(14)=ttoffset
c  criteria for accuracy of path integration (1.e-7 suggested)
c      par(15)=epsilon

c  yes, no bottom reflection flag
c      par(16)=ibot
c  tolerance for bottom reflection
c      par(17)=btol
c  maximum number of bottom reflections allowed before dropping ray
c      par(18)=MBONK
c  total number of bathymetry data points
c      par(19)=numbot
c  maximum depths for rays
c      par(20)=XMAXD
       
c  number of depths in sound speed data
c      par(21)=izdex
c  number of profiles in sound speed data
c      par(22)=irdex

c   ssp,zz,rr define the sound speed section.
c   bb is a two column matrix defining bathymetry (range, depth).
c   trk is a matrix of rays used as a template for tracking rays.
c   s is the vector of four values defining step size.
c
c   Returns xinfo and rpath arrays of ray information.

c  Brian Dushaw
c  Applied Physics Laboratory 
c  University of Washington
c  Seattle, WA 98105
c  (206) 543-1300
c  dushaw@apl.washington.edu

c  Documentation (e.g., an APL Technical Memorandum) may be found at:
c  http://staff.washington.edu/dushaw/eigenray/index.html
c  http://909ers.apl.washington.edu/twiki/bin/viewauth/Main/EigenRay

c  EIGENRAY Version 3.4
c  September-December 1998
c  Last updated March 2013

c Historical Quote (from Moler and Solomon, 1970) 
c "The complete program is not complicated.  It requires less than 400 FORTRAN 
c  statements, including the differential equation solver, but not including the 
c  plotting system.  The solution of a problem [like finding eigenrays at 24 mile 
c  range] requires only a minute or two of elapsed time at a teletype and costs 
c  less than $10.00 (at the University of Michigan's noncommerical rates)."

      implicit double precision (a-h,o-z)

      parameter (NUM_BINS=500)

      include 'nssp.h'
      include 'table.h'
      include 'path.h'
      include 'nbath.h'
      include 'bathy.h'
      include 'nrays.h'
      include 'constants.h'
      
c The dimensions for the variables passed in from the include files
      double precision ssp(NZSSP,Nprofmax),zz(NZSSP),rr(Nprofmax)
c Define trk and trk0 - save trk to restore trk0 should eig be called 
c multiple times
      double precision s(4),trk(neig,8),trk0(neig,8)
      double precision bb(2,MBATH)
      double precision par(25),xinfo(9,MAXRAYS)
      double precision rpath(3,KMAXX,neig)
      double precision xp(KMAXX),yp(2,KMAXX),deltaX
      integer kount,iprof,kmax

cccccccccccccccccccccccccccccccccccccccccccccccccccccc

      double precision eigenangle(MAXRAYS),ystart(4)
      double precision angle_save1(MAXRAYS,3),angle_save(MAXRAYS,3)
      integer indx(MAXRAYS)
c angle save will store the time front of source angle, receiver depth, and 
c "check" which tells which side of the source depth the ray is.

      integer numrays,numrays1,numrays2,numrays3,num_eig,iraysave
      double precision anglebin(NUM_BINS,2)
      double precision bisec(neig,3),eigenangle0(neig)
      
      kmax=0
      npeig=KMAXX

      XMAXDEPTH=par(20)
      izdex=int(par(21))
      irdex=int(par(22))
c fill in the sound speed table
      call speed(ssp,zz,rr,izdex,irdex)

c these values are within the btable common block and so are available to bathy
      ibot=int(par(16))
      btol=par(17)
      MBONK=int(par(18))
c array bb is dimensioned 2XMBATH
c there are numbot values of bathymetry (numbot <= MBATH)
c this initiates numbot for the common block of bathy.h
      numbot=int(par(19)) 
c fill in the bathymetry table
      if (ibot.eq.1) then
c           print *,numbot, MBATH
           if (numbot.gt.MBATH) stop 'Increase MBATH in bathy.h'
           call bathy(bb)
      end if
c btab, etc. filled in


c  fill the step size table
      call stps(s)
c steps filled in

c     zero rpath
      do j=1,3
         do k=1,npeig
            do m=1,neig
               rpath(j,k,m)=zero
            enddo
         enddo
      enddo

ccccccc
ccc Define the variables needed
ccccccc
      ang1=par(1)
      ang2=par(2)
      numrays1=int(par(3))
      z0=par(4)
      zr=par(5)
      r1=zero
      r2=par(6)

      itimefront=int(par(7))
      itemplate=int(par(8))
      nrays=int(par(9))
      iraysave=int(par(10))
      dxsav=par(11)
      zmiss=par(12)
      zmiss_max=par(13)

      ttoffset=par(14)
      epsilon=par(15)

      if (itemplate.eq.1) then
c  trk0 might get overwritten by a previous ray ID
c  so restore the template
         do k=1,8
             do j=1,neig
                 trk0(j,k)=trk(j,k)
c                print *,trk(j,k)
             enddo
         enddo
      endif
      
c  Since most of the rays go to the cusps of the time front, a trick
c  is employed to get more rays in the upper ocean where you want them.
c  ray angles giving arrivals above depth_omit are found from the first
c  numrays1 rays, and numrays2 rays are calculated for those ray angles.
c  If the receiver depth is shallow, we want to at least go down to the
c  sound channel axis, seems to me.
      depth_omit=max(zr*2.0d0,1000d0)
c  Arbitrarily setting numrays2 to be a quarter of the number of requested rays.
c  So, if 4000 rays are requested, 1000 more rays will be used to better
c  fill in the timefront.
      numrays2=int(numrays1*0.25d0)

ccccccccc
ccc  Should have all the data we need to begin the ray traces
ccccccccc

cccccccccccc  BEGIN
c need to flip sign of angles so they are consistent with code convention.
      ang1=-ang1*torad
      ang2=-ang2*torad
      ang1=dsin(ang1)
      ang2=dsin(ang2)

      ystart1=ang1

      if(numrays1.ne.1)then
        delta=(ang2-ang1)/(numrays1-1)
      else
        delta=zero
      endif

c flat earth transform the source depth
      eps=z0*invRe
      z0=z0*(one+eps*half+eps*eps*third)
c flat earth transform the receiver depth
      eps=zr*invRe
      zr=zr*(one+eps*half+eps*eps*third)

c      NN_OMP=nint(float(numrays1/100)) 
c      print *,'Using ', NN_OMP, ' threads.'
c      call OMP_set_num_threads(NN_OMP)
c      call OMP_set_num_threads(16)

      if (itimefront.eq.1) then
c  save all the angles into array eigenangle to get the time front
c      print *
c      print *, 'Calculating a time front.'
           do i=1,numrays1
             eigenangle(i)=ystart1
             ystart1=ystart1+delta
           enddo
          num_eig=numrays1
      else
c  itimefront=0: so get the eigenrays
c      print *
c      print *, 'Searching for eigenrays.'
c      print *, 'Begin initial sweep of time front.'
c  we are ready to go...


!$OMP PARALLEL 
!$OMP DO PRIVATE(ystart1,ystart,xp,yp,deltaX,kount,iprof,kmax,check)
      do 111 i=1,numrays1
c increment angle - increment sin(angle) actually to avoid calculating sin(theta).
      ystart1=ang1+(i-1)*delta
      ystart(1)=dsqrt(one - ystart1*ystart1)
      ystart(2)=ystart1
      ystart(3)=z0
      ystart(4)=zero

c ystart is replaced by the values the end of the
c integration by odeint.  Thus, this will be angle at receiver and
c depth of ray at receiver.  ystart1 is the (sine of) source angle
      
      call odeint(ystart,r1,r2,xp,yp,deltaX,kount,iprof,kmax)

c ystart(2) is now sin(ray angle) at receiver
c ystart(3) is now ray depth at receiver

      check=dsign(one,zr-ystart(3))

      angle_save1(i,1)=ystart1
      angle_save1(i,2)=ystart(3)
      angle_save1(i,3)=check

111   enddo
!$OMP END DO
!$OMP END PARALLEL
      icount=numrays1

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cccccc  End of first sweep through the time front   ccccccccccccccccccccccccccccccc
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cccccc  Now find the angle bins needed to focus on arrivals above depth_omit  ccccc
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

        do i=1,NUM_BINS
         anglebin(i,1)=99.d0
         anglebin(i,2)=99.d0
        enddo

       anglebin(1,1)=ang1
       numbins=1
       ibinsave=-5

       do 112 i=2,numrays1

c depth is positive down
        if (angle_save1(i-1,2).gt.depth_omit
     *.and. angle_save1(i,2).lt.depth_omit) then
           anglebin(numbins,1)=angle_save1(i-1,1)
           ibinsave=i-1
        elseif (angle_save1(i-1,2).lt.depth_omit
     *.and. angle_save1(i,2).gt.depth_omit
     *.and. abs(ibinsave-i).ge. 5 ) then
c          Do the above to get rid of rays that might "jitter" across depth_omit
c          We are only looking at the upper parts of the time front, which go from
c          up to down.  On the down part, we want to be sure that we are more than
c          5 rays away from the up part, i.e. we are not just jittering back and forth
c          across depth_omit on the up part.
           anglebin(numbins,2)=angle_save1(i,1)
           numbins=numbins+1
        endif

112     enddo

        if(anglebin(numbins,1).ne.99.d0
     *.and.anglebin(numbins,2).eq.99.d0)then
            anglebin(numbins,2)=ang2
            numbins=numbins+1
        endif
c we will always have at least one bin, even if numrays1=0

       numbins=numbins-1
       numrays=numrays2/numbins
c trace numrays in each bin, irrespective of the width of the bin.

c now trace the rays in each bin, saving the launch angle, receiver depth
c and check value.
      do 66 kkk=1,numbins

      delta=xpi*(anglebin(kkk,2)-anglebin(kkk,1))/numrays
c xpi is an "irrational" number slightly less than 1 (pi/3.142 actually)
c so that the rays this second time around will not exactly reproduce any 
c of the rays the first time around. 
!$OMP PARALLEL 
!$OMP DO PRIVATE(ystart1,ystart,xp,yp,deltaX,kount,iprof,kmax,check)
      do 113 i=1,numrays
c increment angle (or sin(angle), rather).
      ystart1=anglebin(kkk,1)+i*delta
      ystart(1)=dsqrt(one - ystart1*ystart1)
      ystart(2)=ystart1
      ystart(3)=z0
      ystart(4)=zero

      call odeint(ystart,r1,r2,xp,yp,deltaX,kount,iprof,kmax)
c ystart(2) is now sin(ray angle) at receiver
c ystart(3) is now ray depth at receiver

      check=dsign(one,zr-ystart(3))
 
      angle_save1(numrays1+(kkk-1)*numrays+i,1)=ystart1
      angle_save1(numrays1+(kkk-1)*numrays+i,2)=ystart(3)
      angle_save1(numrays1+(kkk-1)*numrays+i,3)=check
 113  enddo
!$OMP END DO
!$OMP END PARALLEL 

c end of loop over bins below
 66   enddo
      icount=icount+numbins*numrays

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cccc  End of second sweep through the timefront   cccccccccccccccccccccccccc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c  now sort angle_save1 and save to angle_save
c  use the array eigenangle() as a temporary array
c  these must be ordered largest to smallest, note opposite angle sign convention
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      do i=1,icount
        eigenangle(i)=-angle_save1(i,1)
      enddo

      call indexx(icount,eigenangle,indx)
c which returns with the index array indx

      do i=1,icount
          do j=1,3
         angle_save(i,j)=angle_save1(indx(i),j)
          enddo
      enddo

      numrays3=icount

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
cccc  Have now completed filling in the timefront as best as possible  ccccc
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      print *,'Now beginning search for eigenrays.'
      print *
c Now redo the loop and search for eigenrays.  It is set up this way because
c it is better for parallel computation.  And perhaps even for searching for
c missed eigenrays (not implemented yet.)

c We will have saved numrays*numbins = numrays2 (within 1 or 2) rays

        num_eig=0
        check=angle_save(1,3)
        kk=1
        do 114 k=2,numrays3
           kk=kk+1
           if(kk.gt.numrays3) goto 115
           iprev=kk-1

c a badly traced ray is flagged with a -1 depth = angle_save(iprev,2)
c this happens when a ray gets trapped in a surface mixed layer, for example.
97        if(angle_save(iprev,2).lt.zero.and.iprev.gt.2) then
                iprev=iprev-1
                goto 97
           endif

          check0=angle_save(kk,3)
        if ((check0.ne.check).and.(angle_save(kk,2).ge.zero).and.
     1(check0.ne.angle_save(iprev,3)))then
c check that angle_save(j,3)[or ystart(2)].ge.0 above because we dont 
c want to consider unphysical rays.
c if check0 is equal to angle_save(iprev,3) (and check0 is not equal to check)
c then we have probably incremented kk too far or not enough.  Ignore it and
c dont change the value of check.

c first check to see if the end points happen to be at reciever depth
c this happens often enough to be worthwhile.
          iflag=0

          dmiss=dabs(angle_save(iprev,2)-zr)
          if(dmiss.le.zmiss)then
            num_eig=num_eig+1
            eigenangle(num_eig)=angle_save(iprev,1)
            iflag=1
          endif

          dmiss=dabs(angle_save(kk,2)-zr)
          if(dmiss.le.zmiss)then
            num_eig=num_eig+1
            eigenangle(num_eig)=angle_save(kk,1)
            iflag=1
          endif

          if(iflag.eq.0)then
c We have crossed reciever depth; find eigenray by bisection.
c Just record the relevant parameters for now.
            ystart1=angle_save(kk,1) 
            delta=angle_save(kk,1)-angle_save(iprev,1)
            num_eig=num_eig+1
            bisec(num_eig,1)=ystart1
            bisec(num_eig,2)=delta
            bisec(num_eig,3)=check0
            eigenangle(num_eig)=9999.
c flag it for calculation later
          endif

c we will presume that we can skip 5 rays before getting to the next eigenray
c this will cut down on the duplicated rays.
c      kk=kk+5 ...let us not assume this anymore - a better integration
      check=-check

        endif
114   enddo
115   continue

!$OMP PARALLEL 
!$OMP DO PRIVATE(ystart1,delta,check0,eiga,isuccess)
      do 116 kk=1,num_eig
        if (eigenangle(kk).gt.999.0) then
         ystart1=bisec(kk,1)
         delta=bisec(kk,2)
         check0=bisec(kk,3)
         call find(r1,r2,z0,zr,check0,zmiss,ystart1,delta,eiga,isuccess)
c Call find (find_arr.f) to get the eigenray, reports back isuccess=0 if unsucessful.
         if(isuccess.eq.1)then
c N.B. isuccess is not always 1
             eigenangle(kk)=eiga
         endif
        endif
116   enddo
!$OMP END DO
!$OMP END PARALLEL 

c gyrations to clean up for isuccess not always 1
      nnn=0
      do 117 kk=1,num_eig
             if (eigenangle(kk).lt.999.0) then
                nnn=nnn+1
                eigenangle0(nnn)=eigenangle(kk)
             endif
117   enddo

      num_eig=nnn
      do 118 kk=1,num_eig
         eigenangle(kk)=eigenangle0(kk)
118   enddo
c End of attempt to find the eigenray angles.

      print *, 'Final calculation of eigenray parameters and ray paths'
      print *

c endif for itimefront check
      endif
c*******************************************************************************
c now ready for final ray trace of eigenrays.
c all of the eigenray launch angles are in the array eigenangle(num_eig)

c   save raypaths this time around - even if we dont write them out we still need to
c   get the turning point information.      
       kmax=KMAXX

c      print *,'%   j   time      init.angle arr.angle low.turn    u
c     1p.turn  z rec.   s sp. #tp'   
c107   format(a) 

      if (num_eig .gt. neig .and. itimefront.eq.0) then
        print *,'WARNING:  More eigenrays found than allocated.'
        print *,'Maximum eigenrays: ',neig
        print *,'Found ',num_eig,' eigenrays'
        print *,'Only calculating first ',neig,' rays to avoid a crash.'
        print *,'Adjust input parameters, or change neig and recompile.'
        num_eig=neig
      endif

!$OMP PARALLEL 
!$OMP DO PRIVATE(ystart,xp,yp,deltaX,kount,iprof,kmax,
!$OMP1 dmiss,ang,recang,zrec,crec,ttt,eps,fac2,
!$OMP1 ntpt,nUtp,nLtp,ztlow,ztup,igoingup,iindex)
      do 300 i=1,num_eig
       kmax=KMAXX
c eigenangle(i) has the sine of the eigenangles, actually. 
c Get cos(theta) by simple relation to sin(theta).
       ystart(1)=dsqrt(one - eigenangle(i)*eigenangle(i))
       ystart(2)=eigenangle(i)
       ystart(3)=z0
       ystart(4)=zero

      call odeint(ystart,r1,r2,xp,yp,deltaX,kount,iprof,kmax)
c      print *,i,ystart(1),ystart(2),ystart(3),ystart(4),r1,r2

       dmiss=dabs(ystart(3)-zr)
       if(dmiss.gt.zmiss_max.and.itimefront.ne.1) goto 299
c ray is outside of what we will REALLY tolerate

c  The ray paths are stored in variables xp(KMAXX) and yp(2,KMAXX)

c  xp(KMAXX) is range r

c  yp(1,*) is depth of ray z(r), and yp(2,*) is ds/c^2

c ystart(1) is now cos(ray angle) at receiver
c ystart(2) is now sin(ray angle) at receiver
c ystart(3) is now ray depth at receiver
c ystart(4) is now travel time to receier

       ang= -todeg*dasin(eigenangle(i))
       recang=-todeg*dasin(ystart(2))
       zrec=ystart(3)
       if(zrec.lt.zero) goto 299

       call getc(zrec,crec,deltaX,iprof)

       ttt=ystart(4)-ttoffset

c unstretch the depths of the rays...
      do k=1,kount
       if (k.gt.1.and.xp(k).lt.one) goto 199
         eps=yp(1,k)*invRe
         fac2=one/(one+eps*half+eps*eps*third)
         yp(1,k)=yp(1,k)*fac2
      enddo

199    ntpt=0
       nUtp=0
       nLtp=0
       ztlow=zero
       ztup=6000.0
       if(yp(1,5).gt.yp(1,1))then
c  we are going down, so igoingup = 0
          igoingup=0
       else
          igoingup=1
       endif
c  Now count the number of turning points and Upper and Lower turning depths 
c  compare k with k-5 to avoid small fluctuations that may occur right at the 
c  turning point, though there is no evidence that this problem might occur.
       iindex=6
       do k=6,kount
         if (iindex.gt.kount) goto 213
         if (xp(iindex).lt.one) goto 213
         ztlow=dmax1(ztlow,yp(1,iindex))
         ztup=dmin1(ztup,yp(1,iindex))

         if(igoingup.eq.0.and.yp(1,iindex).lt.yp(1,iindex-5))then
           nLtp=nLtp+1
           ntpt=ntpt+1
           igoingup=1
           iindex=iindex+10
         elseif(igoingup.eq.1.and.yp(1,iindex).gt.yp(1,iindex-5))then
           nUtp=nUtp+1
           ntpt=ntpt+1
           igoingup=0
           iindex=iindex+10
         endif

         iindex=iindex+1
       enddo

213    continue
 
         xinfo(1,i)=float(i)
         xinfo(2,i)=ttt
         xinfo(3,i)=ang
         xinfo(4,i)=recang
         xinfo(5,i)=ztlow
         xinfo(6,i)=ztup
         xinfo(7,i)=zrec
         xinfo(8,i)=crec
         xinfo(9,i)=float(ntpt)
c       write(6,100)icount,ttt,ang,recang,ztlow,ztup,zrec,crec,ntpt
c100   format(i5,1x,f12.7,2x,f8.4,4x,f8.4,4x,f6.0,2x,f6.0,3x,f12.6,1x,
c     1f7.2,i4)

      if(iraysave.eq.1)then
c Fill in the ray path array, rpath, to send back to the driver program.
        do k=1,kount
          if (k.gt.1.and.xp(k).lt.one) goto 99
          rpath(1,k,i)=xp(k)
          rpath(2,k,i)=-yp(1,k)
          rpath(3,k,i)=yp(2,k+1)
          if(k.eq.kount) rpath(3,k,i)=zero
        enddo
99      continue
      endif
       
299   continue
300   enddo
!$OMP END DO
!$OMP END PARALLEL

c Move the id part of the code out of the parallel loop - there
c may have been parallelization issues with the template.
      if(itemplate.ne.0) then
         do 301 i=1,num_eig
           ttt=xinfo(2,i)
           ang=xinfo(3,i)
           recang=xinfo(4,i)
           ztlow=xinfo(5,i)
           ztup=xinfo(6,i)
           zrec=xinfo(7,i)
           crec=xinfo(8,i)
           ntpt=int(xinfo(9,i))

           call findrays(trk0,ttt,ang,recang,ntpt,nrays,itemplate)
c            findrays is in the code rayID.f
c  trk0 template is initiated by the driver to eig.f
c  Returns itemplate = 9 if the traced ray was not identified with one of the template rays.
c  Returns itemplate = 2 if the ray is identified.
c  If no template is available, then itempl=0.  Just forget about it; all the ray results 
c     (eigenrays or otherwise) will return.
       
            if(itemplate.eq.9)then
c the ray was not identified - zero it out.
              xinfo(2,i)=zero
              xinfo(3,i)=zero
              xinfo(4,i)=zero
              xinfo(5,i)=zero
              xinfo(6,i)=zero
              xinfo(7,i)=zero
              xinfo(8,i)=zero
              xinfo(9,i)=zero
            endif
301      enddo
      endif

      icount=num_eig
   
c  save the number of rays found, and the total number of rays traced
c  (e.g., finding 100 eigenrays might require 5000 rays)
      par(23)=icount
      par(24)=iraycount

      return
      end
