      subroutine speed(ssp,zz0,rr,ndep,nprof)
c
c This subroutine uses cubic splines for sound speed interpolation.
c This subroutine uses the sound speed map ssp(sspz, range) to make a 
c look up table etab of cubic spline coefficients.
c Several arrays are already defined in the 'table.h' include file.
c This routine is called only once; see eigenray.f
c
c This routine is written with an acknowledgement to Jim Bowlin's
c RAY code and the look up table method he derived. (alas, it is in c...)
c
c For now it assumes that the user has made an intelligent set of sound
c speed profiles.  Peculiar or absent values of sound speed will cause
c the cubic splines to behave badly and produce erroneous results.  
c
c Adds another value for sound speed just above the surface for insurance
c purposes, ie. for surface reflected rays that end up reflecting an
c infinitesimal bit above the surface.  
c
c  Revised 2/08 to use netlib routines cubspl and ppvalu from pppack 
c  and "not a knot" boundary conditions.  Does two splines - the first
c  to spline c onto a closely sampled set of depths, the second to
c  calculate spline coeeficients on those closely sampled set of depths.
c  With the spline coeeficients, c, dc/dz can be easily calculated.
c
c  Brian Dushaw
c  Applied Physics Laboratory 
c  University of Washington
c  Seattle, WA 98105
c  (206) 543-1300
c  dushaw@apl.washington.edu
c
c  EIGENRAY Version 3.4
c  September-December 1998
c  Last updated November 2012

      implicit double precision (a-h,o-z)
      include 'nssp.h'
      include 'table.h'
      include 'constants.h'
c cry is a table of sound speed and coefficients determined by the call to cubspl.
      double precision zz(NZSSP),zz0(NZSSP),cry(4,NZSSP)
      double precision ssp(NZSSP,Nprofmax),rr(Nprofmax)
c tab2 and cry2 are local variables with indexes 1:Ndepmax+2.
c They are used to interface with the spline routines which
c won't deal with indexes -1:Ndepmax.
      double precision tab2(Ndepmax+2),cry2(4,Ndepmax+2)

c Flag to print out sound speeds; change to 1 to get a print out.  
c Data read in goes to std out.
c Splined profiles go to "c_out.dat".
c N.B.:  Turns off flat earth correction as well.
      iprintflag=0
      if (iprintflag.eq.1) open(56,file='c_out.dat')

      if (XMAXDEPTH.eq.zero) XMAXDEPTH=Ndepmax*depth_inc

c tab(i) is the depth at index i - fill it in.
      do 3 i=-1,Ndepmax
          tab(i)=dble(i)*depth_inc
          tab2(i+2)=tab(i)
3     enddo

      Ndep1=ndep+1
      do 65 k=1,nprof
         range(k)=rr(k)
         do j=1,ndep
            eps=zz0(j)*invRe
            if (iprintflag.eq.1) eps=zero
            cry(1,j+1)=ssp(j,k)*(one+eps+eps*eps)
            do ik=2,4
              cry(ik,j+1)=zero
            enddo
            if (k.eq.1) then
              zz(j+1)=zz0(j)*(one+eps*half+eps*eps*third)
            endif
c Apply flat earth transformation.  This correction is inherent
c in the differential equations with the spherical transformation, but applying
c this transformation now saves computation time.  Derivs.f can now use the
c cartesian ODE's.  Z remains positive down.
         enddo 

c Add another point above the surface in case the ray runs
c over the surface by a tiny amount.  Add some sound speed to make
c the ray go back into the water; don't want zero gradient or to
c get the spline over excited.
         zz(1) = -depth_inc
         cry(1,1) = cry(1,2) + 0.2d0*depth_inc
         do ik=2,4
            cry(ik,1)=zero
         enddo

         if (iprintflag.eq.1)then
           do j=1,Ndep1
              print *,k,zz(j),cry(1,j)
           enddo
         endif

         call cubspl(zz,cry,Ndep1,0,0)
c Returns coefficients for obtaining sound speed, dc/dz and d2c/dz2 
c These things are calculated by the ppvalu function
c Evaluate sound speed at all the small depth intervals.
         do 64 jj=-1,Ndepmax
           ztemp=tab2(jj+2)
           ctemp=ppvalu(zz,cry,Ndep1-1,4,ztemp,0)
           if (iprintflag.eq.1) write(56,*) k,ztemp,ctemp
           cry2(1,jj+2)=ctemp
           do ik=2,4
             cry2(ik,jj+2)=zero
           enddo
64       enddo

c Fill etab - cubic spline coefficients on the depth_inc-m grid
c This is a cubic spline of a cubic spline, true.
         call cubspl(tab2,cry2,Ndepmax+2,0,0)
         do jj=-1,Ndepmax
        if (iprintflag.eq.1)write(56,*) k+1000,tab2(jj+2),cry2(1,jj+2) 
           do ik=1,4
              etab(ik,jj,k)=cry2(ik,jj+2)
           enddo
         enddo

c end do for nprof loop
65    enddo
     
      if (iprintflag.eq.1) close(56)

c Now all the sound speed parameters relating to the vertical 
c are filled in.  Get the horizontal range differences:
c These are just pre-calculated for efficiency.
      if(nprof.gt.1)then
         do k=1,nprof-1
             xinvr(k)=one/(range(k+1)-range(k))
         enddo
c For nprof, use a previous value, this will never be used anyways
         xinvr(nprof)=xinvr(nprof-1)
      else
c     Want no range dependence.
         xinvr(1)=0.0d0
      endif

c  Repeat the last profile at a very large range for insurance purposes.  
      range(nprof+1)=1.d9
      do jj=-1,Ndepmax
         do ik=1,4
           etab(ik,jj,nprof+1)=cry2(ik,jj+2)
         enddo
      enddo

      return
      end
